%    This file is part of the psPack simulation package from the
%    doctoral dissertation "A DAE Formulation for Multi-Zone
%    Application to CVCP Engines" by Per Öberg. 
%    The dissertation is available at 
%    http://urn.kb.se/resolve?urn=urn:nbn:se:liu:diva-18120
%
%    Copyright (C) 2013  Per Öberg <per.oberg@liu.se>
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License along
%    with this program; if not, write to the Free Software Foundation, Inc.,
%    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

function psCycleSelect(block)

    setup(block);

%endfunction

%% Function: setup ===================================================
%% Abstract:
%%   Set up the S-function block's basic characteristics such as:
%%   - Input ports
%%   - Output ports
%%   - Dialog parameters
%%   - Options
%% 
%%   Required         : Yes
%%   C-Mex counterpart: mdlInitializeSizes
%%
function setup(block)
% disp(sprintf('Running psCycleSelect:Setup at time: %d',NaN))
  % Register number of ports
  block.NumInputPorts = 0; 
  block.NumOutputPorts = 3; 
  block.NumContStates = 0; 
  % Setup port properties to be inherited or dynamic
  %  block.SetPreCompPortInfoToDefaults;
  
  % Override output port properties
  block.OutputPort(1).DatatypeID  = 0; % double
  block.OutputPort(1).Complexity  = 'Real';

  % Register parameters
  block.NumDialogPrms     = 1;

  % Register sample times
  %  [0 offset]            : Continuous sample time
  %  [positive_num offset] : Discrete sample time
  %
  %  [-1, 0]               : Port-based sample time
  %  [-2, 0]               : Variable sample time
  block.SampleTimes = [-2 0];
  
  block.OutputPort(1).Dimensions  = 1;
  block.OutputPort(2).Dimensions  = 1;
  block.OutputPort(3).Dimensions  = 1;

  % Never accept frame data... 
  block.OutputPort(1).SamplingMode  = 0;
  block.OutputPort(2).SamplingMode  = 0;
  block.OutputPort(3).SamplingMode  = 0;

  
  %% -----------------------------------------------------------------
  %% Register methods called during update diagram/compilation
  %% -----------------------------------------------------------------
  block.RegBlockMethod('PostPropagationSetup', @DoPostPropSetup);
  
  %% -----------------------------------------------------------------
  %% Register methods called at run-time
  %% -----------------------------------------------------------------
  
  block.RegBlockMethod('InitializeConditions', @InitializeConditions);
  block.RegBlockMethod('Start', @Start);
  block.RegBlockMethod('Outputs', @Outputs);

  %% 
  %% Update:
  %%   Functionality    : Called to update discrete states
  %%                      during simulation step
  %%   C-Mex counterpart: mdlUpdate
  %%
  block.RegBlockMethod('Update', @Update);
  
  

%endfunction

%% -------------------------------------------------------------------
%% The local functions below are provided for illustrative purposes
%% to show how you may implement the various block methods listed
%% above.
%% -------------------------------------------------------------------


function InitializeConditions(block)
% disp(sprintf('Running psCycleSelect:Initialize at time: %d',block.currentTime))
  
    global psSimStruct

    %warning('psCycleSelect:InitCond','You must set up a correct initialization!');
    % Tänk igenom det här ordentligt. Tänk på att update sker efter
    % output i varje steg. Dvs saker händer i följande ordning:
    % 1) Setup
    % 2) SetInPortFrameData
    % 3) DoPostPropSetup
    % 4) Start (time exists from hereon)
    % 5) Init 
    % 6) Output
    % 7) Update
    % 8) Repeat 6-7 and possibly 5-7
    % Start kan alltså anvädas första ggn modellen startas för att
    % avgöra vars vi är  medans initialize kan köras varje gång vi kommer
    % tillbaks från ett disablat tillstånd.

    caOffset = block.DialogPrm(1).Data; 
    currentCa = mod(block.currentTime*psSimStruct.psSimPar.w+caOffset,4*pi);
    currentCycleNbr = floor((block.currentTime*psSimStruct.psSimPar.w+ ...
                             caOffset)/4/pi); 
    
    
    %% Get the events that are forthcoming, excluding events that we know
    % have passed even though there might be rounding errors.
    lastCycleNbr = currentCycleNbr;
    lastEventCa = currentCa;
    events = psGetEvents(currentCa,lastCycleNbr,lastEventCa);
    
    
    % Eftersom funktionen i inialize körs precis före output, och inte som 
    % i update fallet efter output så skall alltså initialize ställa in
    % värden som svarar mot föregående update. Dvs den update som skulle ha
    % skett tidigare. 
    %
    % För val av event betyder det att vi som next event skall välja det
    % event som är närmast föregående. Vi kan då inte använda samma
    % output metod som tidigare eftersom den bygger på att funktionen
    % endast körs i några specifika event-punkter.
    % 
    
    %    currentCa-events.eventCa(end)
    
    %    (currentCa-caOffset+4*pi*currentCycleNbr)/psSimStruct.psSimPar.w-(events.eventCa(1)-caOffset+4*pi*currentCycleNbr)/psSimStruct.psSimPar.w
    
    nextEventCa = events.eventCa(end);
 
    block.Dwork(2).data = nextEventCa;
    % Save last cycleNbr used
    block.Dwork(3).data = lastCycleNbr;
    

    if nextEventCa == events.EVO || nextEventCa == events.EVC || nextEventCa == events.IVO
        block.Dwork(1).Data = 1;
    elseif nextEventCa == events.IVC;
        block.Dwork(1).Data = 2;
    elseif nextEventCa == events.SOC;
        block.Dwork(1).Data = 3;
    end
    
%endfunction

function Start(block)
  %disp(sprintf('Running psCycleSelect:Start at time: %d',block.currentTime))
%endfunction

function DoPostPropSetup(block)
  %disp(sprintf('Running psCycleSelect:DoPostPropSetup at time: %d',NaN))
  block.NumContStates = 0;
  block.NumDworks = 3;
  
  block.Dwork(1).Name            = 'CyclePart';
  block.Dwork(1).Dimensions      = 1;
  block.Dwork(1).DatatypeID      = 0;      % double
  block.Dwork(1).Complexity      = 'Real'; % real
  block.Dwork(1).UsedAsDiscState = true;

  block.Dwork(2).Name            = 'lastEventCa';
  block.Dwork(2).Dimensions      = 1;
  block.Dwork(2).DatatypeID      = 0;      % double
  block.Dwork(2).Complexity      = 'Real'; % real
  block.Dwork(2).UsedAsDiscState = true;

  block.Dwork(3).Name            = 'lastCycleNbr';
  block.Dwork(3).Dimensions      = 1;
  block.Dwork(3).DatatypeID      = 0;      % double
  block.Dwork(3).Complexity      = 'Real'; % real
  block.Dwork(3).UsedAsDiscState = true;

  %% Register all tunable parameters as runtime parameters.
  block.AutoRegRuntimePrms;

%endfunction


function Outputs(block)
% disp(sprintf('Running psCycleSelect:Output at time: %d',block.currentTime))
    if block.Dwork(1).Data == 1
        block.OutputPort(1).Data = 1; 
        block.OutputPort(2).Data = 0; 
    else
        block.OutputPort(1).Data = 0; 
        block.OutputPort(2).Data = 1; 
    end
    block.OutputPort(3).Data = block.Dwork(1).Data; 
%endfunction

function Update(block)
% disp(sprintf('Running psCycleSelect:Update at time: %d',block.currentTime))

    global psSimStruct;

    caOffset = block.DialogPrm(1).Data;
    currentCa = mod(block.currentTime*psSimStruct.psSimPar.w+caOffset,4*pi);
    %% Get the forthcoming events (in Order) for the following lap,
    %% excluding events that we know have passed event though there might
    %% be rounding errors. 
    lastCycleNbr = block.Dwork(3).Data;
    lastEventCa = block.Dwork(2).Data;
    events = psGetEvents(currentCa,lastCycleNbr,lastEventCa);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Calculate the nex time hit.
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    nextEventCa = events.eventCa(1);
    if  nextEventCa < lastEventCa
        % No more events this cycle...
        nextCycleNbr = lastCycleNbr+1;
    else
        nextCycleNbr = lastCycleNbr;
    end
    nextEventTime = (nextCycleNbr*4*pi+events.eventCa(1)-caOffset)/psSimStruct.psSimPar.w;
    %disp(sprintf('currentCA is %f',currentCa))
    %disp(sprintf('Next eventCA %f',nextEventCa))
    %events

    block.NextTimeHit =nextEventTime;
    % Save last eventTime used for checking next cycle
    block.Dwork(2).data = nextEventCa;
    % Save last cycleNbr used
    block.Dwork(3).data = nextCycleNbr;        
    
    if nextEventCa == events.EVO;
        block.Dwork(1).Data = 1;
    elseif nextEventCa == events.IVC;
        block.Dwork(1).Data = 2;
    elseif nextEventCa == events.SOC;
        block.Dwork(1).Data = 3;
    elseif nextEventCa == events.EOC;
        block.Dwork(1).Data = 3;
    end


%     if nextEventCa == events.EVO;
%         disp('EVO is going to happen')
%     elseif nextEventCa == events.IVC;
%         disp('IVC is going to happen')
%     elseif nextEventCa == events.SOC;
%         disp('SOC is going to happen')
%     elseif nextEventCa == events.IVO;
%         disp('IVO is going to happen')
%     elseif nextEventCa == events.EVC;
%         disp('EVC is going to happen')
%     end



%endfunction

